"""Mostly context agnostic helpers"""
import sys
import atexit
from threading import Event
import curses

FINISHED = Event()


def end_curse():
    """Clean up our curses environment, ignoring any
    problems  while doing so."""
    try:
        curses.endwin()
    except curses.error:
        pass


def problems(msg):
    """Rudimentary handling of problems"""
    end_curse()
    print("Problem: %s" % msg)
    FINISHED.set()
    sys.exit(1)


def do_write(data, dest):
    """Cleanly write a file out somewhere"""
    handle = open(dest, 'w')
    handle.write(data)
    handle.close()


def finish():
    """Clean up and signal that the end has arrived to
    our various Vault polling threads"""
    end_curse()
    FINISHED.set()


atexit.register(finish)



"""CLI Entrypoint"""

import os
import sys
from getpass import getpass
import logging
import cryptorito
from propriecle import gui
from propriecle.helpers import problems, do_write
from propriecle.servers import get_server
from propriecle.vault import unseal, init, seal, step_down, rekey_start, \
    rekey_enter, rekey_cancel, regenerate_start, regenerate_enter, \
    regenerate_cancel
from propriecle.keys import list_keys, get_root_token, grok_keys
from propriecle.filez import root_file_name, unseal_file_name
import propriecle.conf as conf
LOG = logging.getLogger(__name__)

def cli_root(name):
    """Prints the decrypted root token to stdout"""
    root = get_root_token(get_server(name))
    if root:
        print(root)
    else:
        problems("Root token unavailable")


def cli_unseal(name):
    """Attempts to submit every available unseal key to
    the specified Vault instance"""
    server = get_server(name)
    client = server['client']
    for key_obj in [k for k in list_keys(server) if k['key']]:
        unseal(client, key_obj['key'])


def cli_unseal_all():
    """Attempts to submit every available unseal key for each
    Vault instance in order"""

    for name in [x['name'] for x in conf.get('vaults')]:
        cli_unseal(name)


def cli_seal_all():
    """Seals every accessible Vault instance"""
    for name in [x['name'] for x in conf.get('vaults')]:
        seal(get_server(name))


def cli_init(name):
    """Initializes Vault on the specified instance"""
    server = get_server(name)
    if not init(server):
        sys.exit(1)

    sys.exit(0)


def cli_step_down(name):
    """Asks the specified Vault instance to step down from Leader"""
    step_down(get_server(name))
    sys.exit(0)


def cli_seal(name):
    """Seals the specified Vault instance"""
    if not seal(get_server(name)):
        sys.exit(1)

    sys.exit(0)


def cli_regenerate_start(name):
    """Start the root key regeneration process"""
    regenerate_start(get_server(name))


def cli_regenerate_auth(name):
    """Attempts to submit every available unseal key in support
    of the root key regeneration process"""
    server = get_server(name)
    for key_obj in [k for k in list_keys(server) if k['key']]:
        regenerate_enter(server, key_obj['key'])


def cli_regenerate_cancel(name):
    """Cancels the root key regeneration process"""
    if not regenerate_cancel(get_server(name)):
        sys.exit(1)

    sys.exit(0)


def cli_rekey_start(name):
    """Start the unseal rekey process"""
    rekey_start(get_server(name), grok_keys())


def cli_rekey_auth(name):
    """Attempts to submit every available unseal key in support
    of the unseal rekey process"""
    server = get_server(name)
    for key_obj in [k for k in list_keys(server) if k['key']]:
        rekey_enter(server, key_obj['key'])


def cli_rekey_cancel(name):
    """Cancels the unseal rekey process"""
    if not rekey_cancel(get_server(name)):
        sys.exit(1)

    sys.exit(0)


def cli_root_import(name):
    """Imports a plaintext root token and will encrypt
    according to the propriecle configuration."""
    server = get_server(name)
    root_token = getpass('Root Token: ', stream=sys.stderr)
    if not root_token:
        problems("Must specify a token")

    root_key = conf.get('root_key')
    key_id = cryptorito.key_from_keybase(root_key[8:])['fingerprint']
    encrypted = cryptorito.portable_b64encode(cryptorito.encrypt_var(root_token, [key_id]))
    do_write(encrypted, root_file_name(server))

def cli_unseal_import(name, s_slot):
    """Imports a unseal key at a spcified slot and will
    encrypt accordign to the propriecle configuration."""
    slot = int(s_slot)
    server = get_server(name)
    unseal_key = getpass('Unseal Key: ', stream=sys.stderr)
    if not unseal_key:
        problems("Must specify a unseal key")

    a_key = conf.get('keys')[slot - 1]
    key_id = cryptorito.key_from_keybase(a_key[8:])['fingerprint']
    encrypted = cryptorito.portable_b64encode(cryptorito.encrypt_var(unseal_key, [key_id]))
    do_write(encrypted, unseal_file_name(server, slot))


def main():
    """Entrypoint Actual"""
    if os.environ.get('PROPRIECLE_LOG'):
        logging.basicConfig(level=logging.DEBUG)

    root_log = logging.getLogger()
    logfile = "%s/propriecle.log" % os.environ.get('PROPRIECLE_LOG_PATH', os.getcwd())
    if os.path.exists(logfile):
        os.remove(logfile)

    file_handler = logging.FileHandler(logfile)
    if len(sys.argv) == 1:
        root_log.handlers = []
        root_log.addHandler(file_handler)
        LOG.debug("with %s handlers", len(root_log.handlers))
        gui()
    elif len(sys.argv) == 3 and sys.argv[1] == "unseal":
        cli_unseal(sys.argv[2])
    elif len(sys.argv) == 2 and sys.argv[1] == "unseal":
        cli_unseal_all()
    elif len(sys.argv) == 3 and sys.argv[1] == "seal":
        cli_seal(sys.argv[2])
    elif len(sys.argv) == 2 and sys.argv[1] == "seal":
        cli_seal_all()
    elif len(sys.argv) == 3 and sys.argv[1] == "init":
        cli_init(sys.argv[2])
    elif len(sys.argv) == 3 and sys.argv[1] == "step_down":
        cli_step_down(sys.argv[2])
    elif len(sys.argv) == 3 and sys.argv[1] == "root_get":
        cli_root(sys.argv[2])
    elif len(sys.argv) == 3 and sys.argv[1] == "rekey_start":
        cli_rekey_start(sys.argv[2])
    elif len(sys.argv) == 3 and sys.argv[1] == "rekey_auth":
        cli_rekey_auth(sys.argv[2])
    elif len(sys.argv) == 3 and sys.argv[1] == "rekey_cancel":
        cli_rekey_cancel(sys.argv[2])
    elif len(sys.argv) == 3 and sys.argv[1] == "regenerate_start":
        cli_regenerate_start(sys.argv[2])
    elif len(sys.argv) == 3 and sys.argv[1] == "regenerate_auth":
        cli_regenerate_auth(sys.argv[2])
    elif len(sys.argv) == 3 and sys.argv[1] == "regenerate_cancel":
        cli_regenerate_cancel(sys.argv[2])
    elif len(sys.argv) == 3 and sys.argv[1] == 'root_import':
        cli_root_import(sys.argv[2])
    elif len(sys.argv) == 4 and sys.argv[1] == 'unseal_import':
        cli_unseal_import(sys.argv[2], sys.argv[3])
    else:
        sys.exit(1)



"""Low level Vault interactions"""
import os
import requests
import hvac
import hvac.exceptions
from propriecle.helpers import do_write
from propriecle.filez import root_file_name, unseal_file_name
from propriecle.keys import get_root_token, grok_key, grok_keys
import propriecle.conf as conf


def init(server):
    """Initialize a fresh Vault server. It will do this based on the
    settings as defined in the proprieclerc file"""
    obj = {
        "root_token_pgp_key": grok_key(conf.get('root_key')),
        "secret_shares": len(conf.get('keys')),
        "secret_threshold": conf.get('required'),
        "pgp_keys": grok_keys()
    }
    client = server['client']
    resp = client.write('sys/init', **obj)

    if 'root_token' not in resp or \
       'keys_base64' not in resp:
        return False

    do_write(resp['root_token'], root_file_name(server))
    index = 0
    for key in resp['keys_base64']:
        do_write(key, unseal_file_name(server, index + 1))
        index = index + 1

    return True


def unseal(client, key):
    """Unseals a Vault server with the provided (decrypted) key"""
    client.unseal(key)


def seal(server):
    """Seals the Vault server. Requires the use of a root token"""
    client = root_client(server)
    if client:
        # can't seal a leader but you can if it's not ha
        if server['leader'] or not server['ha']:
            client.seal()

        return True
    else:
        return False


def root_client(server):
    """This is like a normal Vault client but with the root token instead"""
    client = server['client']
    root_token = get_root_token(server)
    if not root_token:
        return
    else:
        setattr(client, 'token', root_token)

    if not am_root(client):
        return None

    return client


def am_root(client):
    """Determines if a vault client is root or not"""
    try:
        token = client.lookup_token()
    except (hvac.exceptions.VaultDown, requests.exceptions.ReadTimeout):
        return False

    if 'data' in token and 'policies' in token['data']:
        return 'root' in token['data']['policies']

    return False


def get_vault(server):
    """Manifests an hvac Vault Client object"""
    verify_tho = server.get('tls_verify', True)
    if not verify_tho:  # :(
        import urllib3
        urllib3.disable_warnings()

    client = hvac.Client(server['url'],
                         timeout=5,
                         verify=verify_tho)

    token_file = "%s/.vault-token" % os.environ['HOME']
    if 'VAULT_TOKEN' in os.environ and\
       os.environ['VAULT_TOKEN']:
        setattr(client, 'token', os.environ['VAULT_TOKEN'])
    elif os.path.exists(token_file):
        handle = open(token_file, 'r')
        setattr(client, 'token', handle.read().strip())
        handle.close()

    return client


def rekey_start(server, keys):
    """Start the rekeying process. This will result in new
    unseal keys being generated once quorum has been achieved."""
    client = server['client']
    required = conf.get('required')
    backup = conf.get('backup', False)
    client.start_rekey(secret_shares=len(keys),
                       secret_threshold=required,
                       backup=backup,
                       pgp_keys=keys)


def rotate_master(server):
    """Rotate the master key. This requires a unsealed Vault server
    and a root (or equivalent) token"""
    client = root_client(server)
    if client:
        client.rotate()
        return True

    return False


def regenerate_start(server):
    """Begin the process for regenerating the root token"""
    client = server['client']
    obj = {
        "pgp_key": grok_key(conf.get('root_key'))
    }
    client.write('sys/generate-root/attempt', **obj)


def step_down(server):
    """Ask this Vault instance to step down. Is a noop if the
    instance is already in standby"""
    client = root_client(server)
    client.write('sys/step-down')


def regenerate_enter(server, key):
    """Submit a unseal key in support of regenerating the root
    token. Returns true once the operation has completed."""
    client = server['client']

    regen_obj = client.read('sys/generate-root/attempt')
    obj = {
        'key': key,
        'nonce': regen_obj['nonce']
    }
    try:
        resp = client.write("sys/generate-root/update", **obj)
    except hvac.exceptions.InvalidRequest as vault_exception:
        # eventual consistency fallout
        if vault_exception.errors[0] == 'no root generation in progress':
            return True

    if 'complete' not in resp:
        return False

    if resp['complete']:
        root_file = root_file_name(server)
        do_write(resp['encoded_root_token'], root_file)

    return True


def regenerate_cancel(server):
    """Cancels the regeneration of the root token"""
    client = server['client']
    client.delete('sys/generate-root/attempt')
    return True


def rekey_enter(server, key):
    """Submit a unseal key in support of rekeying the unseal keys"""
    client = server['client']
    rekey_obj = client.rekey_status
    resp = client.rekey(key, rekey_obj['nonce'])
    if 'complete' not in resp:
        return False

    if resp['complete']:
        index = 0
        for b_key in resp['keys_base64']:
            unseal_file = unseal_file_name(server, index + 1)
            do_write(b_key, unseal_file)
            index = index + 1

        return True

    return False


def rekey_cancel(server):
    """Cancel the request to rekey the unseal keys"""
    client = server['client']
    client.cancel_rekey()
    return True



"""Handle interactions with our local (hopefully) encrypted files"""
from hashlib import sha256
import logging
import propriecle.conf as conf
from propriecle.conf import get_server
LOG = logging.getLogger(__name__)

def root_file_name(server):
    """The filename of our root token"""
    filename = "%s/%s-root" % (conf.state_directory(),
                               server_hash(server))
    return filename


def unseal_file_name(server, suffix):
    """The filename of a suffixed unseal key"""
    filename = "%s/%s-unseal.%s" % (conf.state_directory(),
                                    server_hash(server),
                                    suffix)
    return filename


def server_hash(server):
    """Generate a hash of a server (or it's static parent)"""
    if not isinstance(server, dict):
        return server_hash(get_server(server))

    a_server = server
    if 'parent' in server:
        LOG.debug("has a parent %s", a_server['parent'])
        a_server = get_server(a_server['parent'])

    LOG.debug("has_server is %s", a_server)
    return sha256(a_server['url']).hexdigest()



"""All dat conf"""

import os
import yaml
from propriecle.helpers import problems


def config_file():
    """Returns the config file or a potentially sane default"""
    if 'PROPRIECLE_CONFIG' in os.environ:
        return os.environ['PROPRIECLE_CONFIG']

    return "%s/.proprieclerc" % os.environ['HOME']


class PropriecleConfig(object):
    """A class representing the runtime configuration"""
    config = {}

    def __init__(self):
        filename = config_file()
        if os.path.exists(filename):
            self.config = yaml.safe_load(open(filename).read())
        else:
            problems("No config file found")

    def get(self, key, default):
        return self.config.get(key, default)

MY_CONFIG = PropriecleConfig()


def get(key, default=None):
    """Return a config entry, or the default if nothing explicitly
    set has been found"""
    return MY_CONFIG.get(key, default)


def state_directory():
    """Return the state directory. This is used for saving encrypted keys."""
    prefix = "%s/.propriecle" % os.environ['HOME']
    if 'PROPRIECLE_DIRECTORY' in os.environ:
        prefix = os.environ['PROPRIECLE_DIRECTORY']

    return prefix


def get_server(name):
    """Returns the basic configuration expression of
    a Vault server. This is the basis for the full expression of
    the server but lacks details."""
    if isinstance(name, dict):
        return get_server(name['name'])

    vaults = get('vaults')
    for s in vaults:
        if name == s['name']:
            return s

    problems("Invalid Vault")



"""All the UI Widgets we need, and probably more"""

import curses
import time


def init_screen():
    """Initialize our base curses constructs. Sets up the screen object, sets
    our default color pairs, and sets basic terminal options."""
    screen = curses.initscr()
    curses.start_color()
    curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)  # default
    curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)  # OK
    curses.init_pair(3, curses.COLOR_RED, curses.COLOR_BLACK)  # NOT OK
    curses.init_pair(4, curses.COLOR_WHITE, curses.COLOR_RED)  # alart
    curses.init_pair(5, curses.COLOR_GREEN, curses.COLOR_RED)  # alart effect
    curses.init_pair(6, curses.COLOR_WHITE, curses.COLOR_BLUE)  # selection
    curses.init_pair(7, curses.COLOR_YELLOW, curses.COLOR_BLACK)  # selection

    # do not display curses, or echo anything
    curses.curs_set(0)
    curses.cbreak()
    curses.noecho()
    # capture keys!
    screen.keypad(1)
    return screen


def popup(screen, msg):
    """Generate a popup message that requires a acknoledgement"""
    middle_y, middle_x = middle_yx(screen)
    my_x = len(msg) + 6
    my_y = 3
    win = curses.newwin(my_y, my_x, middle_y - my_y, middle_x - my_x)
    screen.refresh()
    win.border()
    win.bkgd(ord(' '), curses.color_pair(1))
    string_at(win, 1, 0, msg, curses.color_pair(7), 'center')
    start = time.time()
    while True:
        win.refresh()
        screen.timeout(500)
        if screen.getch() != -1 or time.time() - start > 2:
            win.erase()
            win.refresh()
            del win
            return


def yesno(screen, msg):
    """Generate a popup message that a user can accept or cancel"""
    middle_y, middle_x = middle_yx(screen)
    my_x = len(msg) + 6
    my_y = 4
    win = curses.newwin(my_y, my_x, middle_y - my_y, middle_x - my_x)
    screen.refresh()

    win.border()
    win.bkgd(ord(' '), curses.color_pair(1))
    string_at(win, 1, 0, msg, curses.color_pair(2), 'center')
    done = False
    blink = False
    msg_yes = '[Y]es'
    msg_no = '[N]o'
    anim_x = (my_x/2) + (len(msg)/2)
    while not done:
        if not blink:
            string_at(win, 1, anim_x, '?', curses.color_pair(2))
            string_at(win, 2, 2, msg_yes,
                      curses.color_pair(2) | curses.A_BOLD)
            string_at(win, 2, -2, msg_no, curses.color_pair(2), 'right')
            blink = True
        else:
            string_at(win, 1, anim_x, ' ', curses.color_pair(2))
            string_at(win, 2, 2, msg_yes, curses.color_pair(2))
            string_at(win, 2, -2, msg_no,
                      curses.color_pair(2) | curses.A_BOLD, 'right')
            blink = False

        win.refresh()
        screen.timeout(500)
        ch = screen.getch()
        if ch == ord('y') or ch == ord('Y'):
            pulse(win, 2, 2, msg_yes, 'left')
            del win
            return True
        elif ch == ord('n') or ch == ord('N') or ch == 27:
            pulse(win, 2, -2, msg_no, 'right')
            del win
            return False


def pulse(win, y, x, msg, align):
    """Generates some prettily pulsing text"""
    for f in range(1, 6):
        if f % 2 == 0:
            string_at(win, y, x, msg, curses.color_pair(2), align)
        else:
            string_at(win, y, x, msg, curses.color_pair(3), align)

        win.refresh()
        curses.napms(200)


def middle_yx(screen):
    """Determins the middle of the screen"""
    xy = screen.getmaxyx()
    screen_y = xy[0]
    screen_x = xy[1]
    return screen_y/2, screen_x/2


def string_at(win, y, x_base, msg, attr, align='left'):
    """Places a string at a particular place on the screen."""
    x = win.getmaxyx()[1]
    str_x = 1
    if align == 'center':
        str_x = (x/2)-(len(msg)/2)
    elif align == 'right':
        str_x = x - len(msg) - 1

    str_x = str_x + x_base

    win.addstr(y, str_x, msg, attr)


def ask_for(screen, msg):
    """Text input funciton. Pretty terrible, tbh."""
    middle_y, middle_x = middle_yx(screen)
    window_width = len(msg) + 10
    win_x = middle_x - (window_width/2)
    win_y = middle_y - 10
    win = curses.newwin(4, window_width, win_y, win_x)
    screen.refresh()
    win.border()
    win.bkgd(ord(' '), curses.color_pair(1))
    win.addstr(1, (window_width / 2) - (len(msg) / 2),
               msg, curses.color_pair(2))
    resp = ''
    done = False
    step = 0
    steps = ['/', '|', '\\', '-']
    while not done:
        win.refresh()
        screen.timeout(500)
        ch = screen.getch()
        if ch == 27:
            win.erase()
            win.refresh()
            del win
            return None
        elif ch == 10:
            win.erase()
            win.refresh()
            del win
            done = True
        else:
            if ch != -1:
                resp = "%s%s" % (resp, chr(ch))
            else:
                win.addch(1,
                          (window_width/2) + (len(msg)/2) + 2,
                          ord(steps[step]),
                          curses.color_pair(2) | curses.A_BOLD)
                if step == len(steps) - 1:
                    step = 0
                else:
                    step = step + 1

    return resp



"""Maintain a local cache of Vault server details"""
import threading
import requests
from propriecle.helpers import FINISHED
from propriecle.vault import get_vault, root_client
from propriecle.conf import get_server as server_from_config
from future.utils import iteritems  # pylint: disable=E0401
import hvac.exceptions

VTHREADZ = []
VAULTZ = {}
SERVERZ = {}


def update_server(finished, server):
    """Updates our local poorly cached representation of a Vault server.
    This runs in a thread, one per Vault server. Probably needs more mutex."""
    name = server['name']
    while not finished.wait(2):
        new_s = fetch_server(VAULTZ[name], server)
        if 'cluster_id' in new_s:
            my_cluster = [x['name']
                          for _name, x
                          in iteritems(SERVERZ)
                          if x.get('cluster_id', None) == new_s['cluster_id']]
            new_s['cluster_members'] = my_cluster

        SERVERZ[name] = new_s


def get_server(name):
    """Will return the cached rerepsentation of a Vault server or if this isn't
    available will kick off a thread to start that up"""
    if name in SERVERZ:
        return SERVERZ[name]

    server = server_from_config(name)
    return start_server_thread(server)


def start_server_thread(server):
    """Starts a thread responsible for updating local info on Vault servers"""
    client = None
    name = server['name']
    if name in VAULTZ:
        client = VAULTZ[name]
    else:
        client = VAULTZ[name] = get_vault(server)

    if name in SERVERZ:
        return SERVERZ[name]

    server = SERVERZ[name] = fetch_server(client, server)
    sthread = threading.Thread(target=update_server,
                               args=(FINISHED, server,))
    VTHREADZ.append(sthread)
    sthread.setDaemon(True)
    sthread.start()
    return server


def fetch_server(client, server):
    """Fetches details from Vault, generating a complicate dict
    representing a server"""
    server_obj = {
        "name": server['name'],
        "url": server['url'],
        "client": client
    }
    if (server['name'] in SERVERZ) \
       and ('cluster_members' in SERVERZ[server['name']]):
        clustered = SERVERZ[server['name']]['cluster_members']
        server_obj['cluster_members'] = clustered

    if 'parent' in server:
        server_obj['parent'] = server['parent']

    try:
        init = client.is_initialized()
    except hvac.exceptions.InternalServerError:
        return server_obj
    except requests.ConnectionError:
        return server_obj
    except requests.ReadTimeout:
        return server_obj

    server_obj['init'] = init
    if init:
        try:
            status = client.seal_status
        except (hvac.exceptions.VaultDown, requests.exceptions.ReadTimeout):
            return server_obj

        server_obj['version'] = status['version']
        seal = server_obj['sealed'] = status['sealed']
        if seal:
            server_obj['unseal_required'] = status['t']
            server_obj['unseal_progress'] = status['progress']
        else:
            server_obj['cluster_name'] = status['cluster_name']
            server_obj['cluster_id'] = status['cluster_id']

            server_obj['rekey'] = False
            try:
                rekey_obj = client.rekey_status
                rekey = server_obj['rekey'] = rekey_obj['started']
                if rekey:
                    server_obj['rekey_backup'] = rekey_obj['backup']
                    server_obj['rekey_progress'] = rekey_obj['progress']
                    server_obj['rekey_required'] = rekey_obj['required']
            except (hvac.exceptions.VaultDown, requests.exceptions.ReadTimeout):
                pass
            except hvac.exceptions.InternalServerError as vault_exception:
                if vault_exception.message == 'node not active but active '\
                   'node not found':
                    pass

            server_obj['ha'] = False
            server_obj['leader'] = False
            try:
                leader_obj = client.read('sys/leader')
                server_obj['ha'] = leader_obj['ha_enabled']
                if leader_obj['ha_enabled']:
                    server_obj['leader'] = leader_obj['is_self']
            except (hvac.exceptions.VaultDown, requests.exceptions.ReadTimeout):
                pass
            except hvac.exceptions.InternalServerError as e:
                if e.message == 'node not active but active node not found':
                    pass

            if not server_obj['ha'] or \
               (server_obj['ha'] and server_obj['leader']):
                try:
                    regen_obj = client.read('sys/generate-root/attempt')
                    server_obj['regenerating'] = regen_obj.get('started', False)
                    if server_obj['regenerating']:
                        server_obj['regen_progress'] = regen_obj['progress']
                        server_obj['regen_required'] = regen_obj['required']
                except (hvac.exceptions.VaultDown, requests.exceptions.ReadTimeout):
                    pass
                except hvac.exceptions.InternalServerError as e:
                    if e.message == 'node not active but ' \
                       'active node not found':
                        pass

            client = root_client(server_obj)
            server_obj['is_root'] = False
            if client:
                try:
                    key_obj = client.key_status
                    server_obj['key_term'] = key_obj['term']
                    server_obj['is_root'] = True
                except (hvac.exceptions.VaultDown, requests.exceptions.ReadTimeout):
                    pass
                except hvac.exceptions.InternalServerError as e:
                    if e.message == 'node not active but active '\
                       'node not found':
                        pass

    return server_obj



"""Propriecle grab bag"""
import sys
from random import SystemRandom
import curses
import logging
import traceback
import hvac
from propriecle.ui import string_at, middle_yx, yesno, \
    popup, init_screen, ask_for
from propriecle.helpers import problems, finish
from propriecle.vault import seal, unseal, \
    rekey_start, rotate_master, regenerate_start, init, \
    regenerate_enter, regenerate_cancel, rekey_enter, rekey_cancel, \
    step_down
from propriecle.keys import list_keys, grok_keys
from propriecle.servers import get_server
import propriecle.conf as conf
LOG = logging.getLogger(__name__)

def update_servers():
    """Kicks off the update process for our Vault servers"""
    servers = []
    for server in conf.get('vaults'):
        servers.append(get_server(server['name']))

    return servers


def do_seal(screen, server):
    """Invoke the GUI action of sealing Vault"""
    name = server['name']

    seal_msg = "SEAL %s" % name
    if not yesno(screen, seal_msg):
        return
    screen.refresh()

    try:
        if not seal(server):
            popup(screen, "Unable to seal!")
    except hvac.exceptions.Forbidden:
        popup(screen, "Invalid root token!")
        return


def do_rekey(server):
    """Invoke the GUI action of rekeying unseal keys"""
    if not conf.get('keys') or not conf.get('required'):
        problems('keys and required missing from config')

    rekey_start(server, grok_keys())


def get_keys(screen, server):
    """Generates a list of keys for the user to select from.
    Options also include manually entering a key or using
    any valid key"""
    client = server['client']
    keys = list_keys(server)
    middle_y, middle_x = middle_yx(screen)
    my_x = 30
    my_y = len(keys) + 6
    win_x = middle_x - (my_x/2)
    win_y = middle_y - 10
    win = curses.newwin(my_y, my_x, win_y, win_x)
    status = client.seal_status

    done = False
    selection = 0
    return_keys = []
    while not done:
        win.border()
        win.bkgd(ord(' '), curses.color_pair(1))
        win.addstr(1, 1, "Select Key (%s of %s complete)" %
                   (status['progress'], status['t']),
                   curses.color_pair(2))
        win.addstr(my_y - 2, 1, "[m]anual [a]ny", curses.color_pair(1))
        for key_obj in keys:
            name = key_obj['name']
            i = key_obj['index']
            color = curses.color_pair(2)
            if not key_obj['key']:
                color = curses.color_pair(4)

            if selection == i:
                color = curses.color_pair(6)

            if name.startswith('keybase:'):
                name = name[8:]

            win.addstr(i + 2, 1, "[%s] %s" % (i + 1, name), color)

        win.refresh()
        screen.timeout(500)
        ch = screen.getch()
        if ch == 27:
            del win
            screen.refresh()
            return
        elif ch == curses.KEY_DOWN:
            down_done = False
            while not down_done:
                if selection < len(keys) - 1 and \
                   keys[selection + 1]['key']:
                    selection = selection + 1
                    down_done = True
                elif selection == len(keys) - 1:
                    down_done = True
                else:
                    selection = selection + 1
        elif ch == curses.KEY_UP:
            up_done = False
            while not up_done:
                if selection > 0 and \
                   keys[selection - 1]['key']:
                    selection = selection - 1
                    up_done = True
                elif selection == 0:
                    up_done = True
                else:
                    selection = selection - 1
        elif ch == ord('m') or ch == ord('M'):
            return ask_for(screen, "Key")
        elif ch == ord('a') or ch == ord('A'):
            return_keys = [k for k in keys if k['key']][0:conf.get('required')]
            done = True
        elif ch == 10:
            if keys[selection]['key']:
                done = True

    win.erase()
    win.refresh()
    del win

    if len(return_keys) == 0:
        return_keys = [k for k in keys if k['index'] == selection]

    return return_keys


def do_unseal(screen, server):
    """Invoke the GUI action of unsealing a Vault server"""
    client = server['client']
    status = client.seal_status
    if not status['sealed']:
        popup(screen, "Not actually sealed")
        return

    for key_obj in do_get_keys(screen, server):
        try:
            unseal(client, key_obj['key'])
        except hvac.exceptions.InvalidRequest:
            popup(screen, "Unable to unseal with %s:%s" %
                  (key_obj['index'], key_obj['name']))
            popup(screen, "Invalid Key!")


def refresh_screen(screen, selection, main_win=None):
    """Refreshes the GUI elements of the main screen"""
    y, x = screen.getmaxyx()
    main_x = x - (x / 4)
    main_y = len(conf.get('vaults')) + 5
    if y < main_y or x < main_x:
        problems("screen too small")

    screen.border()
    screen.bkgd(ord(' '), curses.color_pair(1))

    middle_y, middle_x = middle_yx(screen)
    if not main_win:
        main_win = curses.newwin(main_y, main_x,
                                 int(round(middle_y - (main_y/2))),
                                 int(round(middle_x - (main_x/2))))
    else:
        main_win.erase()

    main_win.bkgd(ord(' '), curses.color_pair(1))
    main_win.border()
    string_at(main_win, 1, 0, conf.get("title", "Vault Friends"),
              curses.color_pair(1), 'center')
    refresh_status(main_win, selection)
    screen.nooutrefresh()
    main_win.nooutrefresh()
    curses.doupdate()
    return main_win


def refresh_initialized(main_win, server, index, selected):
    """Refreshes the lable for a initialized Vault instance"""
    y, x = main_win.getmaxyx()
    name = server['name']
    if server['init'] and not server['sealed'] and server['ha']:
        c_state = 'S'
        if server['leader']:
            c_state = 'L'

        name = "%s [%s %s]" % (server['name'], c_state, server['cluster_name'])

    name_msg = "[%s] %s" % (index, name)
    base_x = len(name_msg) + 1
    line = index + 3
    col = curses.color_pair(1)
    if selected:
        col = curses.color_pair(6)

    string_at(main_win, line, 0, name_msg, col)
    if server['sealed']:
        s_msg = "Sealed"
        string_at(main_win, line, base_x, s_msg, curses.color_pair(3))
    else:
        s_msg = 'OK'
        s_col = 2
        if server['rekey']:
            s_msg = 'REKEY'
            s_col = 7

        string_at(main_win, line, base_x, s_msg, curses.color_pair(s_col))


def randcolor():
    """A random pretty color plz"""
    s_color = 0
    s_rand = SystemRandom().randint(0, 2)
    if s_rand == 0:
        s_color = 1
    elif s_rand == 1:
        s_color = 3
    elif s_rand == 2:
        s_color = 7

    if SystemRandom().randint(0, 1) == 1:
        s_color = s_color | curses.A_BOLD

    return s_color


def refresh_uninitialized(main_win, server, selected, index):
    """Refreshes the label for a uninitialized Vault instance"""
    y, x = main_win.getmaxyx()
    name = server['name']
    name_msg = "[%s] %s" % (index, name)
    base_x = len(name_msg) + 1
    new_msg = 'NEW'
    line = index + 3
    col = curses.color_pair(1)
    if selected:
        col = curses.color_pair(6)

    string_at(main_win, line, 0, name_msg, col)
    string_at(main_win, line,
              base_x, new_msg,
              curses.color_pair(7) | curses.A_BOLD)
    s_anim = ['*', 'o', '.', '^', '_', '-', '|']

    for f in range(0, 5):
        string_at(main_win,
                  line,
                  base_x + len(new_msg) + 2 + f,
                  s_anim[SystemRandom().randint(0, len(s_anim) - 1)],
                  curses.color_pair(randcolor()))


def refresh_offline(main_win, server, selected, index):
    """Refreshes the label for a offline Vault instance"""
    y, x = main_win.getmaxyx()
    name = server['name']
    name_msg = "[%s] %s" % (index, name)
    base_x = len(name_msg) + 1
    new_msg = 'AWOL'
    line = index + 3
    col = curses.color_pair(1)
    if selected:
        col = curses.color_pair(6)

    string_at(main_win, line, 0, name_msg, col)
    string_at(main_win, line, base_x,
              new_msg, curses.color_pair(3) | curses.A_BOLD)


def refresh_status(main_win, selection):
    """Refreshes the label of a Vault instance"""
    y, x = main_win.getmaxyx()
    index = 0
    for server in update_servers():
        if 'init' in server:
            if server['init']:
                refresh_initialized(main_win, server,
                                    index, selection == index)
            else:
                refresh_uninitialized(main_win, server,
                                      selection == index, index)
        else:
            refresh_offline(main_win, server,
                            selection == index, index)

        index = index + 1


def focus_loop(screen, index):
    """Main interaction loop when an actual Vault instance is selected
    and we are looking at the detailed view."""
    done = False
    screen.erase()
    while not done:
        server = get_server(conf.get('vaults')[index]['name'])
        refresh_focused(screen, server)
        if focus_input(screen, server):
            screen.erase()
            return


def focus_input(screen, server):
    """User input for when we are in the detailed view of a Vault instance"""
    screen.timeout(500)
    ch = screen.getch()
    if ch == 27:
        return True

    if 'init' in server:
        if 255 > ch > 0:
            ch_s = chr(ch).lower()
            is_init = server['init']
            if is_init:
                is_sealed = server['sealed']
                if not is_sealed:
                    is_rekey = server['rekey']
                    is_regenerating = server['regenerating']
                    is_ha = server['ha']
                    if is_ha:
                        is_leader = server['leader']

            if ch_s == 's' and is_init and not is_sealed:
                if not is_ha or (is_ha and is_leader):
                    do_seal(screen, server)
            elif ch_s == 'u' and is_init and is_sealed:
                do_unseal(screen, server)
            elif ch_s == 'r' and is_init and not is_sealed and not is_rekey:
                do_rekey(server)
            elif ch_s == 'i' and not is_init:
                do_init(screen, server)
            elif ch_s == 'e' and is_init and is_rekey:
                do_rekey_enter(screen, server)
            elif ch_s == 'c' and is_init and is_rekey:
                if not rekey_cancel(server):
                    popup(screen, "Unable to cancel rekey")
            elif ch_s == 'e' and is_init and is_regenerating:
                do_regenerate_enter(screen, server)
            elif ch_s == 'c' and is_init and is_regenerating:
                if not regenerate_cancel(server):
                    popup(screen, "Unable to cancel regeneration")
            elif ch_s == 'o' and is_init and not is_rekey and \
                 not is_regenerating and not is_sealed:
                if not rotate_master(server):
                    popup(screen, "Unable to rotate master")
            elif ch_s == 'g' and is_init and \
                 not is_rekey and \
                 not is_regenerating and \
                 not is_sealed:
                regenerate_start(server)
            elif ch_s == 'p' and is_init and \
                 is_ha and is_leader:
                step_down(server)


def do_get_keys(screen, server):
    """Selecting keys from a screen I guess"""
    keys = get_keys(screen, server)
    screen.nooutrefresh()
    if not keys:
        popup(screen, "Must specify at least one key")
        return []

    return keys


def do_regenerate_enter(screen, server):
    """GUI context for submitting unseal keys when
    regenerating root token"""
    for key_obj in do_get_keys(screen, server):
        if not regenerate_enter(server, key_obj['key']):
            popup(screen, "Unable to regenerate with %s:%s" %
                  (key_obj['index'], key_obj['name']))


def do_rekey_enter(screen, server):
    """GUI context for submitting unseal keys when rekeying"""
    for key_obj in do_get_keys(screen, server):
        if not rekey_enter(server, key_obj['key']):
            popup(screen, "Unable to rekey with %s:%s" %
                  (key_obj['index'], key_obj['name']))


def refresh_focused(screen, server):
    """Redraw the focus/detail view of a Vault instance"""
    win_x = 30 + len(server['url'])
    win_y = 10
    middle_y, middle_x = middle_yx(screen)
    win = curses.newwin(win_y, win_x,
                        middle_y - (win_y / 2),
                        middle_x - (win_x / 2))
    win.border()
    win.bkgd(ord(' '), curses.color_pair(1))
    string_at(win, 1, 0, server['name'], curses.color_pair(1), 'center')
    string_at(win, 2, 0, server['url'], curses.color_pair(1))
    if 'init' in server:
        if server['init']:
            string_at(win, 3, 0, server['version'], curses.color_pair(1))
            if server['sealed']:
                msg_s = 'SEALED'
                string_at(win, 5, 0, msg_s,
                          curses.color_pair(3) | curses.A_BOLD)
                unseal_remaining = server['unseal_required'] \
                                   - server['unseal_progress']
                msg_details = "%s keys required" % (unseal_remaining)
                string_at(win, 5, len(msg_s) + 1,
                          msg_details, curses.color_pair(1))
            else:
                msg_s = ''
                if server['rekey']:
                    msg_s = 'REKEY'
                    string_at(win, 5, 0,
                              msg_s, curses.color_pair(7))
                    rekey_remaining = server['rekey_required'] \
                                      - server['rekey_progress']
                    string_at(win, 6, 0,
                              "%s keys required" % (rekey_remaining),
                              curses.color_pair(1))
                elif (not server['ha'] or (server['ha'] and server['leader'])) \
                     and server['regenerating']:
                    msg_s = 'REGEN'
                    string_at(win, 5, 0,
                              msg_s, curses.color_pair(7))
                    string_at(win, 6, 0,
                              "%s keys required" %
                              (server['regen_required'] - server['regen_progress']),
                              curses.color_pair(1))
                else:
                    msg_s = 'OK'
                    string_at(win, 5, 0, msg_s, curses.color_pair(2))
                    if 'key_term' in server:
                        string_at(win, 6, 0,
                                  "Key Term %s" % server['key_term'],
                                  curses.color_pair(1))

                ha_msg = "HA"
                if server['ha']:
                    string_at(win, 5, len(msg_s) + 1,
                              ha_msg, curses.color_pair(1))
                    if server['leader']:
                        string_at(win, 5,
                                  len(msg_s) + len(ha_msg) +
                                  len(server['cluster_name']) + 3,
                                  "Leader", curses.color_pair(1))
                    else:
                        string_at(win, 5,
                                  len(msg_s) + len(ha_msg) +
                                  len(server['cluster_name']) + 3,
                                  "Standby", curses.color_pair(1))
                else:
                    ha_msg = ''

                string_at(win, 5,
                          len(ha_msg) + len(msg_s) + 2,
                          server['cluster_name'], curses.color_pair(1))
        else:
            string_at(win, 5, 0, "Uninitialized Vault", curses.color_pair(1))
    else:
        string_at(win, 4, 0,
                  'OFFLINE', curses.color_pair(3) | curses.A_BOLD, 'center')

    win.hline(win_y - 3, 1, curses.ACS_HLINE, win_x - 2)
    refresh_controls(win, server)
    win.refresh()


def refresh_controls(win, server):
    """Refresh our visible controls. This varies based on server state
    and authentication"""
    esc_msg = '[ESC] Quit'
    seal_msg = '[s]eal'
    unseal_msg = '[u]nseal'
    init_msg = '[i]nit'
    rekey_msg = '[r]ekey'
    cancel_msg = '[c]ancel'
    enter_msg = '[e]nter'
    rotate_msg = 'r[o]tate'
    regen_msg = "re[g]en"
    step_msg = 'ste[p]'

    line = win.getmaxyx()[0] - 2
    string_at(win, line, 0, esc_msg, curses.color_pair(1))
    next_x = len(esc_msg) + 2
    if 'init' in server:
        if not server['init']:
            string_at(win, line, next_x, init_msg, curses.color_pair(3))
        else:
            if server['sealed']:
                string_at(win, line, next_x, unseal_msg, curses.color_pair(1))
            else:
                if (not server['ha'] or (server['ha'] and server['leader'])) \
                   and (server['rekey'] or server['regenerating']):
                    string_at(win, line, next_x,
                              cancel_msg, curses.color_pair(7))
                    string_at(win, line,
                              next_x + len(cancel_msg) + 1,
                              enter_msg, curses.color_pair(1))
                else:
                    if server['is_root']:
                        if not server['ha'] or \
                           (server['ha'] and server['leader']):
                            string_at(win, line, next_x,
                                      seal_msg, curses.color_pair(3))
                            string_at(win, line,
                                      next_x + len(seal_msg) + 1,
                                      rekey_msg, curses.color_pair(1))
                            string_at(win, line,
                                      next_x + len(seal_msg) +
                                      len(rekey_msg) + 2,
                                      rotate_msg, curses.color_pair(1))
                            string_at(win, line,
                                      next_x + len(seal_msg) + len(rekey_msg) +
                                      len(rotate_msg) + 3,
                                      regen_msg, curses.color_pair(1))

                        if server['ha'] and server['leader']:
                            string_at(win, line,
                                      next_x + len(seal_msg) + len(rekey_msg) +
                                      len(rotate_msg) + len(regen_msg) + 4,
                                      step_msg, curses.color_pair(1))
                    else:
                        string_at(win, line, next_x,
                                  rekey_msg, curses.color_pair(1))
                        string_at(win, line,
                                  next_x + len(rekey_msg) + 2,
                                  regen_msg, curses.color_pair(1))


def do_init(screen, server):
    """GUI construct for initializing a blank Vault instance"""
    if not init(server):
        popup(screen, 'Unexpected return during vault init!')


def main_loop(screen):
    """Main interaction loop."""
    done = False
    selection = 0
    main_win = None
    while not done:
        main_win = refresh_screen(screen, selection, main_win)
        screen.timeout(500)
        ch = screen.getch()
        if ch == 27:
            screen.erase()
            screen.refresh()
            popup(screen, "Thank you for playing")
            curses.endwin()
            return
        elif ch == curses.KEY_DOWN:
            if selection < len(conf.get('vaults')) - 1:
                selection = selection + 1
        elif ch == curses.KEY_UP:
            if selection > 0:
                selection = selection - 1
        elif ch == 10:
            focus_loop(screen, selection)


def gui():
    """It's amazing what qualifies as a GUI these days"""
    update_servers()
    screen = init_screen()
    try:
        main_loop(screen)
        finish()
    except Exception:
        screen.erase()
        screen.move(1, 1)
        screen.refresh()
        curses.endwin()
        etype, val, trace = sys.exc_info()
        traceback.print_exception(etype, val, trace)



#!/usr/bin/env python
import os
import sys
from setuptools import setup
from pkg_resources import resource_filename

# depending on your execution context the version file
# may be located in a different place!
vsn_path = resource_filename(__name__, 'propriecle/version')
if not os.path.exists(vsn_path):
    vsn_path = resource_filename(__name__, 'version')
    if not os.path.exists(vsn_path):
        print("%s is missing" % vsn_path)
        sys.exit(1)


setup(name='propriecle',
      version=open(vsn_path, 'r').read(),
      description='something something vault master keys',
      author='Jonathan Freedman',
      author_email='jonathan.freedman@autodesk.com',
      license='MIT',
      url='https://github.com/autodesk/propriecle',
      install_requires=['PyYAML', 'hvac', 'future', 'cryptorito'],
      include_package_data=True,
      packages=['propriecle'],
      entry_points={
          'console_scripts': ['propriecle = propriecle.cli:main']
      },
      package_data={'propriecle': ['version']}
)



"""Properly access secrets and stuff"""

import os
from cryptorito import decrypt_var, portable_b64decode, \
    export_gpg_key, key_from_keybase
from propriecle.filez import unseal_file_name, root_file_name
import propriecle.conf as conf
from propriecle.conf import get_server


def get_root_token(server):
    """Returns the decrypted root token. Will check both the current
    server and, failing that, other hosts in the same cluster."""
    filename = root_file_name(server)
    if not os.path.exists(filename):
        for a_server in server.get('cluster_members', []):
            a_filename = root_file_name(get_server(a_server))
            if os.path.exists(a_filename):
                return do_decrypt(a_filename)

        return None

    return do_decrypt(filename)


def list_keys(server):
    """Retuns a list of all unseal keys which are available"""
    keys = []
    index = 0
    for key in conf.get('keys'):
        key_obj = {
            'index': index,
            'name': key,
            'key': None
        }
        key_file = unseal_file_name(server, index + 1)

        if os.path.exists(key_file):
            key_obj['key'] = do_decrypt(key_file)
        else:
            for a_server in server.get('cluster_members', []):
                a_filename = unseal_file_name(a_server, index + 1)
                if os.path.exists(a_filename):
                    key_obj['key'] = do_decrypt(a_filename)
                    break

        keys.append(key_obj)
        index = index + 1

    return keys


def do_decrypt(filename):
    """GPG decrypt things in files"""
    if not os.path.exists(filename):
        return None

    handle = open(filename, 'r')
    datas = handle.read().strip()
    handle.close()
    datas_bin = portable_b64decode(datas)
    return decrypt_var(datas_bin)


def grok_key(key):
    """Extracts a single key based on ID (which may actually
    be a keybase id"""
    if key.startswith('keybase:'):
        bits = key_from_keybase(key[8:])['bundle'].split('\n')
        return ''.join(bits[3:len(bits)-2])

    return export_gpg_key(key)


def grok_keys():
    """Returns a list of all extracted keys"""
    keys = []
    for key in conf.get('keys'):
        keys.append(grok_key(key))

    return keys
